/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)io.c	5.6 (Berkeley) 3/1/91
 */
char io_rcsid[] = 
  "$Id: io.c,v 1.10 1997/06/08 20:13:51 dholland Exp $";

/*
 * This file contains the I/O handling and the exchange of 
 * edit characters. This connection itself is established in
 * ctl.c
 */

#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <netdb.h>
#include <time.h>
#include "config.h"
#include "talk.h"
#include "des.h"
#include "blowfish.h"
#include "twofish.h"
#include "cast5.h"
#include "sha1.h"

#define MAX_BLOCK_SIZE	16

#define NO_ALGO		0
#define DES3		1
#define BLOWFISH	2
#define TWOFISH		3
#define CAST5		4
 
extern struct hostent *hp;
extern char *his_name;
static unsigned char *key;
static tripledes_ctx des3_ctx;
static BLOWFISH_context bf_ctx;
static TWOFISH_context tf_ctx;
static CAST5_context cast5_ctx;
static SHA1_CONTEXT sha1_ctx;

static struct st_algo_info {
	int algo;
	int keylen;
	int blocksize;
} algo_info;

void read_config(void);
static void init_crypto(void);
static void do_encrypt(char *from, char *to);
static void do_decrypt(char *from, char *to);

/*
 * The routine to do the actual talking
 */
void
talk(void)
{
	fd_set read_set;
	int nb;
	unsigned char buf[BUFSIZ];
	unsigned char *buf_ptr;
	int i;
	unsigned char plain[MAX_BLOCK_SIZE];
	unsigned char cipher[MAX_BLOCK_SIZE];

	init_crypto();
	if (algo_info.algo)
		message("SECURE connection established");
	else
		message("INSECURE connection established");
	beep();
	wrefresh(curscr);

	current_line = 0;

	/*
	 * Wait on both the other process and standard input
	 */
	for (;;) {
		FD_ZERO(&read_set);
		FD_SET(0, &read_set);
		FD_SET(sockt, &read_set);
		nb = select(sockt+1, &read_set, NULL, NULL, NULL);
		if (nb <= 0) {
			if (errno == EINTR) {
				continue;
			}
			/* panic, we don't know what happened */
			p_error("Unexpected error from select");
			quit(0);
		}
		if (FD_ISSET(sockt, &read_set)) { 
			/* There is data on sockt */
			nb = read(sockt, buf, sizeof(buf));
			if (nb <= 0) {
				message("Connection closed. Exiting");
				quit(0);
			}
			if (algo_info.algo) {
				if (nb < algo_info.blocksize) {
					/* The other side isn't using
					 * encryption.
					 */
					algo_info.algo = NO_ALGO;
					message("Warning, changing to INSECURE mode");
					display(&his_win, buf, nb);
				} else {
					for (buf_ptr = buf;
					     nb >= algo_info.blocksize;
					     nb -= algo_info.blocksize,
					     buf_ptr += algo_info.blocksize) {
						memcpy(cipher, buf_ptr,
						       algo_info.blocksize);
						do_decrypt(cipher, plain);
						display(&his_win, plain, 1);
					}
				}
			} else
				display(&his_win, buf, nb);
		}
		if (FD_ISSET(0, &read_set)) {
			buf[0] = getch();
			nb = 1;
			display(&my_win, buf, nb);
			if (algo_info.algo) {
				plain[0] = buf[0];
				for (i = 1; i < algo_info.blocksize; i++)
					plain[i] = rand();
				do_encrypt(plain, cipher);
				write(sockt, cipher, algo_info.blocksize);
			} else
				write(sockt, buf, nb);
		}
	}
}

/*
 * p_error prints the system error message on the standard location
 * on the screen and then exits. (i.e. a curses version of perror)
 */
void
p_error(const char *string) 
{
	const char *rmcup, *smcup, *amsg="";

	rmcup = tigetstr("rmcup");
	smcup = tigetstr("smcup");

	if (rmcup || smcup) {
		amsg = ". Press any key...";
	}

	wmove(my_win.x_win, current_line%my_win.x_nlines, 0);
	wprintw(my_win.x_win, "[%s : %s (%d)%s]\n",
	    string, strerror(errno), errno, amsg);
	wrefresh(my_win.x_win);
	move(LINES-1, 0);
	refresh();
	if (rmcup || smcup) {
		/* alternative screen, prompt before exit */
		getch();
	}
	quit(0);
}

/*
 * Display string in the standard location
 */
void
message(const char *string)
{

	wmove(my_win.x_win, current_line%my_win.x_nlines, 0);
	wprintw(my_win.x_win, "[%s]\n", string);
	wrefresh(my_win.x_win);
}

/* Reads the ~/.stalkrc file, which contains the passphrases and hostnames
 * to use encryption with
 */
void read_config(void)
{
	FILE *conf;
	int i;
	int pass_len;
	int line_num;
	char conf_file[PATH_MAX];
	char hostname[MAXHOSTNAMELEN];
	char username[NAME_SIZE];
	char line[BUFSIZ];
	char *userhost;
	char *cp;
	char *cp2;
	char *passphrase;
	char *algo;
	struct hostent *remote;
	struct hostent *hp2;
	struct stat file;

	algo_info.algo = NO_ALGO;
	strncpy(conf_file, getenv("HOME"), PATH_MAX);
	strncat(conf_file, "/.stalkrc", PATH_MAX);
	conf_file[PATH_MAX - 1] = '\0';

	if (lstat(conf_file, &file)) {
		perror("stalk");
		fprintf(stderr, "Ignoring ~/.stalkrc\n");
		return;
	} else if (!S_ISREG(file.st_mode)) {
		fprintf(stderr,
			"~/.stalkrc is not a regular file, ignoring.\n");
		return;
	} else if (file.st_mode & ~(S_IFREG | S_IREAD | S_IWRITE)) {
		fprintf(stderr,
			"~/.stalkrc doesn't have secure (600) permissions\n");
		return;
	} else if ((conf = fopen(conf_file, "r")) != NULL) {
		for (line_num = 1; fgets(line, BUFSIZ, conf); line_num++) {
			if (line[0] == '#')
				continue;

			/* Very simple sanity check in the line.
			 * Should be improved.
			 */
			if ((cp = index(line, ' ')) == NULL || line[0] == ' ') {
				fprintf(stderr, ".stalkrc: error in line %d\n",
					line_num);
				continue;
			}
			algo = ++cp;
			if ((cp = index(algo, ' ')) == NULL || algo[0] == ' ') {
				fprintf(stderr, ".stalkrc: error in line %d\n",
					line_num);
				continue;
			}
			*cp = '\0';
			passphrase = ++cp;
			pass_len = strlen(passphrase);
			if (passphrase[pass_len - 1] == '\n')
				passphrase[pass_len - 1] = '\0';

			userhost = index(line, ' ');
			*userhost = '\0';
			userhost = line;
			for (cp = userhost, i = 0; *cp &&
					!index("@:!.", *cp); cp++)
				i++;
			if (*cp == '\0') {
				/* this is a local to local talk */
				strncpy(username, userhost, NAME_SIZE);
				if (gethostname(hostname, MAXHOSTNAMELEN)) {
					perror("stalk");
					break;
				}
			} else {
				if (*cp++ == '@') {
					/* user@host */
					if (i < NAME_SIZE) {
						strncpy(username, userhost, i);
						username[i] = '\0';
					}
					strncpy(hostname, cp, MAXHOSTNAMELEN);
				} else {
					/* host.user or host!user or host:user
					 */
					if (i < MAXHOSTNAMELEN) {
						strncpy(hostname, userhost, i);
						hostname[i] = '\0';
					}
					strncpy(username, cp, NAME_SIZE);
				}
			}

			if (strcmp(username, his_name))
				continue;
			else if ((cp = index(hostname, '*')) != NULL) {
				*cp = '\0';
				if ((hp2 = gethostbyaddr(hp->h_addr_list[0],
							 hp->h_length,
							 hp->h_addrtype)) ==
						NULL) {
					perror("stalkd");
					continue;
				}
				cp2 = &(hp2->h_name[strlen(hp2->h_name) -
						   strlen(++cp)]);
				if (strncmp(cp, cp2, strlen(cp)) ||
				    strncmp(hostname, hp2->h_name,
					    strlen(hostname)))
					continue;

			/* Check which is the right hostname by resolving it
			 * and checking the address with the one we are
			 * connecting with. This can be SLOW, but is used
			 * to make the program work with dynamic IPs.
			 */
			} else if ((remote = gethostbyname(hostname)) == NULL) {
				perror("stalk");
				continue;
			} else if (memcmp((void *) remote->h_addr_list[0],
					(void *) hp->h_addr_list[0],
					remote->h_length))
				continue;

			if (!strcmp(algo, "3des"))
				algo_info.algo = DES3;
			else if (!strcmp(algo, "blowfish"))
				algo_info.algo = BLOWFISH;
			else if (!strcmp(algo, "twofish"))
				algo_info.algo = TWOFISH;
			else if (!strcmp(algo, "cast5"))
				algo_info.algo = CAST5;
			else {
				fprintf(stderr, ".stalkrc: line %d: unknown algorithm\n", line_num);
				break;
			}

			/* Get the hash of the passphrase so we can use
			 * as key for the algorithms
			 */
			sha1_init(&sha1_ctx);
			sha1_write(&sha1_ctx, passphrase, strlen(passphrase));
			sha1_final(&sha1_ctx);
			key = sha1_read(&sha1_ctx);
			break;
		}
		fclose(conf);
	} else {
		perror("stalk");
		fprintf(stderr, "Ignoring ~/.stalkrc\n");
	}
}

void init_crypto(void)
{
	FILE *random;
	unsigned int seed;
	unsigned char *des3_keys[3];

	/* This part is slightly confusing, to deal
	 * with some error conditions...
	 */
	if (!algo_info.algo)
		return;
	if ((random = fopen("/dev/urandom", "r")) == NULL) {
		perror("stalk");
		algo_info.algo = NO_ALGO;
	} else if (!fread(&seed, sizeof(seed), 1, random)) {
		perror("stalk");
		algo_info.algo = NO_ALGO;
	} else {
		srand(seed);
		switch (algo_info.algo) {
			case DES3:
				algo_info.keylen = 8;
				algo_info.blocksize = 8;

				/* Since we have 20 bytes of hash
				 * but need 24, the last key overlaps
				 * the 2nd.
				 */
				des3_keys[0] = &key[0];
				des3_keys[1] = &key[8];
				des3_keys[2] = &key[12];
				tripledes_set3keys(des3_ctx, des3_keys[0],
						   des3_keys[1], des3_keys[2]);
				break;
			case BLOWFISH:
				algo_info.keylen = 16;
				algo_info.blocksize = 8;
				bf_setkey(&bf_ctx, key, 16);
				break;
			case TWOFISH:
				algo_info.keylen = 16;
				algo_info.blocksize = 16;
				twofish_setkey(&tf_ctx, key, 16);
				break;
			case CAST5:
				algo_info.keylen = 16;
				algo_info.blocksize = 8;
				cast_setkey(&cast5_ctx, key, 16);
				break;
		}
	}
	fclose(random);
}

void do_encrypt(char *from, char *to)
{
	switch (algo_info.algo) {
		case DES3:
			tripledes_ecb_encrypt(des3_ctx, from, to);
			break;
		case BLOWFISH:
			bf_encrypt_block(&bf_ctx, to, from);
			break;
		case TWOFISH:
			twofish_encrypt(&tf_ctx, to, from);
			break;
		case CAST5:
			cast_encrypt_block(&cast5_ctx, to, from); 
			break;
	}
}

void do_decrypt(char *from, char *to)
{
	switch (algo_info.algo) {
		case DES3:
			tripledes_ecb_decrypt(des3_ctx, from, to);
			break;
		case BLOWFISH:
			bf_decrypt_block(&bf_ctx, to, from);
			break;
		case TWOFISH:
			twofish_decrypt(&tf_ctx, to, from);
			break;
		case CAST5:
			cast_decrypt_block(&cast5_ctx, to, from);
			break;
	}
}
